//---------------------------------------------------------------------
//  Algorithmic Conjurings @ http://www.coyotegulch.com
//
//  maze_renderer.cpp (libcoyote)
//
//  Maze generation and exploration tools
//-----------------------------------------------------------------------
//
//  COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:
//  
//  If you modify this file, you may insert additional notices
//  immediately following this sentence.
//  
//  Copyright 2001, 2002, 2003 Scott Robert Ladd.
//  All rights reserved, except as noted herein.
//  
//  This computer program source file is supplied "AS IS". Scott Robert
//  Ladd (hereinafter referred to as "Author") disclaims all warranties,
//  expressed or implied, including, without limitation, the warranties
//  of merchantability and of fitness for any purpose. The Author
//  assumes no liability for direct, indirect, incidental, special,
//  exemplary, or consequential damages, which may result from the use
//  of this software, even if advised of the possibility of such damage.
//  
//  The Author hereby grants anyone permission to use, copy, modify, and
//  distribute this source code, or portions hereof, for any purpose,
//  without fee, subject to the following restrictions:
//  
//      1. The origin of this source code must not be misrepresented.
//  
//      2. Altered versions must be plainly marked as such and must not
//         be misrepresented as being the original source.
//  
//      3. This Copyright notice may not be removed or altered from any
//         source or altered source distribution.
//  
//  The Author specifically permits (without fee) and encourages the use
//  of this source code for entertainment, education, or decoration. If
//  you use this source code in a product, acknowledgment is not required
//  but would be appreciated.
//  
//  Acknowledgement:
//      This license is based on the wonderful simple license that
//      accompanies libpng.
//
//-----------------------------------------------------------------------
//  For more information on this software package, please visit
//  Scott's web site, Coyote Gulch Productions, at:
//
//      http://www.coyotegulch.com
//-----------------------------------------------------------------------

#include "maze_renderer.h"
using namespace libcoyote;

#include <stdexcept>
using namespace std;

#define PNG_SETJMP_NOT_SUPPORTED
#include "png.h"

// constructor
maze_renderer::image::image(const maze & source, size_t grid_size)
{
    if (grid_size < 2)
        throw invalid_argument("grid size must be 2 or greater");

    m_width  = static_cast<int>(grid_size * (source.get_width()  + 2));
    m_height = static_cast<int>(grid_size * (source.get_height() + 2));

    size_t alloc_width = (m_width + 7) / 8;

    size_t row, col;

    // allocate row pointers
    m_bits = new unsigned char * [m_height];
    
    // allocate and initialize rows
    for (row = 0; row < size_t(m_height); ++row)
    {
        m_bits[row] = new unsigned char [alloc_width];
        memset(m_bits[row],0xFF,alloc_width);
    }

    // now we set the bits we need to draw
    size_t x = 0, y = 0, x2 = 0, y2 = 0;
    size_t south_row = source.get_height() - 1;
    size_t east_col  = source.get_width() - 1;

    for (col = 0; col <= east_col; ++col)
    {
        maze::cell current_cell;

        x += grid_size;
        y  = 0;

        for (row = 0; row <= south_row; ++row)
        {
            y += grid_size;

            // get the current cell
            current_cell = source.get_cell(col,row);

            // draw west wall
            if (*current_cell.m_walls[maze::DIR_WEST] != maze::WALL_OPEN)
            {
                for (y2 = y; y2 < y + grid_size + 1; ++y2)
                    m_bits[y2][x / 8] &= ~(0x80 >> (x % 8));
            }

            // draw north wall
            if (*current_cell.m_walls[maze::DIR_NORTH] != maze::WALL_OPEN)
            {
                for (x2 = x; x2 < x + grid_size; ++x2)
                    m_bits[y][x2 / 8] &= ~(0x80 >> (x2 % 8));
            }

            // draw south wall when at bottom
            if (row == south_row)
            {
                if (*current_cell.m_walls[maze::DIR_SOUTH] != maze::WALL_OPEN)
                {
                    y2 = y + grid_size;

                    for (x2 = x; x2 < x + grid_size; ++x2)
                        m_bits[y2][x2 / 8] &= ~(0x80 >> (x2 % 8));
                }
            }

            // draw east wall
            if (col == east_col)
            {
                if (*current_cell.m_walls[maze::DIR_EAST] != maze::WALL_OPEN)
                {
                    x2 = x + grid_size;

                    // set bits for this wall segment
                    for (y2 = y; y2 < y + grid_size + 1; ++y2)
                        m_bits[y2][x2 / 8] &= ~(0x80 >> (x2 % 8));
                }
            }
        }
    }
}

// destructor
maze_renderer::image::~image()
{
    // delete image data
    for (int row = 0; row < m_height; ++row)
        delete [] m_bits[row];

    delete [] m_bits;
}

// create a PNG image of the maze in the specified file
void maze_renderer::render(const maze & source, const string & file_name, size_t grid_size)
{
    // libpng requires C-style file I/O
    FILE * file;

    // open the PNG output file
    file = fopen(file_name.c_str(), "wb");

    if (!file)
        throw runtime_error("unable to create PNG file");

    // create a maze image
    image maze_image(source,grid_size);

    // prepare the standard PNG structures
    png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

    if (!png_ptr)
    {
        fclose(file);
        throw runtime_error("failure when calling png_create_write_struct");
    }

    png_infop info_ptr = png_create_info_struct(png_ptr);

    if (!info_ptr)
    {
        fclose(file);
        png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
        throw runtime_error("failure when calling png_create_write_struct");
    }

    // initialize the I/O
    png_init_io(png_ptr, file);
        
    // we're going to write a very simple 3x8 bit RGB image
    png_set_IHDR(png_ptr, info_ptr, maze_image.get_width(), maze_image.get_height(), 1, PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
    
    // write the file header information
    png_write_info(png_ptr, info_ptr);

    // write out the entire image data in one call
    png_write_image (png_ptr, maze_image.get_bits());
    
    // write the additional chunks to the PNG file (not really needed)
    png_write_end(png_ptr, info_ptr);
    
    // clean up after the write, and free any memory allocated
    png_destroy_write_struct(&png_ptr, &info_ptr);

    // close file
    fclose (file);
}

